import datetime
import decimal
import json
import logging
import os
import re
import uuid
from asyncio import Future
from typing import Awaitable, Optional, Union
import ddtrace

from simple_api_framework.monitors import DataDogMonitor

if os.getenv("DD_SERVICE") and os.getenv("DD_API_KEY") and os.getenv("DD_API_KEY") and os.getenv("DD_ENV"):
    ddtrace.patch(tornado=True)
    ddtrace.patch_all()

import dotenv
import magic
import sentry_sdk
from sentry_sdk.integrations.tornado import TornadoIntegration
from tornado import web, ioloop, escape
from tornado.log import enable_pretty_logging

from simple_api_framework.db.mongo import MongoDB
from simple_api_framework.redis import Redis


def setup_logging():
    logger = logging
    enable_pretty_logging()
    return logger


class ValidationException(Exception):
    field = None
    message = None

    def __init__(self, **kwargs):
        for attribute in kwargs.keys():
            if hasattr(self, attribute):
                setattr(self, attribute, kwargs.get(attribute))
        super().__init__()


class ManyValidationExceptions(Exception):
    exceptions = None

    def __init__(self, **kwargs):
        for attribute in kwargs.keys():
            if hasattr(self, attribute):
                setattr(self, attribute, kwargs.get(attribute))
        super().__init__()


class NotFoundHandler(web.RequestHandler):
    MESSAGE = {
        'title': 'Not Found',
        'status': 404,
    }

    async def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    async def get(self):
        self.MESSAGE['instance'] = self.request.path
        self.set_header('Content-type', 'application/problem+json')
        self.set_status(404)
        return self.finish(self.MESSAGE)

    async def post(self):
        self.MESSAGE['instance'] = self.request.path
        self.set_header('Content-type', 'application/problem+json')
        self.set_status(404)
        return self.finish(self.MESSAGE)

    async def put(self):
        self.MESSAGE['instance'] = self.request.path
        self.set_header('Content-type', 'application/problem+json')
        self.set_status(404)
        return self.finish(self.MESSAGE)

    async def patch(self):
        self.MESSAGE['instance'] = self.request.path
        self.set_header('Content-type', 'application/problem+json')
        self.set_status(404)
        return self.finish(self.MESSAGE)

    async def delete(self):
        self.MESSAGE['instance'] = self.request.path
        self.set_header('Content-type', 'application/problem+json')
        self.set_status(404)
        return self.finish(self.MESSAGE)

    async def options(self):
        self.MESSAGE['instance'] = self.request.path
        self.set_header('Content-type', 'application/problem+json')
        self.set_status(404)
        return self.finish(self.MESSAGE)


class Service(web.Application):
    WORKING_DIRECTORY = os.path.dirname(os.path.realpath(__file__)) + '/../'
    NAME = None
    ENVIRONMENT = None

    redis = None
    mongo = None
    monitor = None

    def __init__(self, **kwargs):
        dotenv.load_dotenv(override=True)
        self.logging = setup_logging()
        self.io_loop = ioloop.IOLoop.instance()
        self.ENVIRONMENT = os.getenv('ENV')

        queues = kwargs.get('queues', [])

        urls = kwargs.get('urls')
        if urls:
            if not isinstance(urls, list):
                raise TypeError('Variable "urls" must be a list')
            for url in urls:
                if not isinstance(url, dict):
                    raise TypeError('All items in "urls" must be a dictionaries')

        if not hasattr(self, 'NAME') or not self.NAME:
            raise ValueError('No NAME attribute')

        version = os.getenv("API_VERSION", 1)

        url_prefix = f'/api/v{version}/{self.NAME.lower()}'
        uuid_format = "([a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12})"

        urls.append({'url': 'ping', 'handler': HealthCheckEndpoint})

        handlers = []
        for url in urls:
            uri = url.get('url')
            if '{uuid}' in uri:
                uri = uri.format(uuid=uuid_format)
            if not url.get('without_prefix'):
                handlers.append(web.url(re.compile(f'{url_prefix}/{uri}(/.*)?$'), url.get("handler")))
            else:
                handlers.append(web.url(re.compile(f'{uri}$'), url.get('handler')))

        del kwargs['urls']

        if kwargs.get('static'):
            static = kwargs.get('static')
            handlers.append(
                (
                    f"{url_prefix}/{static.get('url')}",
                    web.StaticFileHandler,
                    {'path': f"{self.WORKING_DIRECTORY}{static.get('path')}"})
            )
            del kwargs['static']

        handlers.append((r"/.*", NotFoundHandler))

        super().__init__(handlers=handlers, debug=True if self.ENVIRONMENT in ['local', 'dev'] else False,
                         xsrf_cookies=False, logging=self.logging, **kwargs)

        if os.getenv("REDIS_URL"):
            self.logging.info("[PLUGINS] Redis enabled")
            self.redis = Redis(prefix=os.getenv("REDIS_PREFIX"), url=os.getenv("REDIS_URL"))

        if os.getenv("MONGODB_URL"):
            self.logging.info("[PLUGINS] MongoDB enabled")
            self.mongo = MongoDB(url=os.getenv("MONGODB_URL"))

        if os.getenv('SENTRY_DSN_URL'):
            self.logging.info("[PLUGINS] Sentry enabled")
            sentry_sdk.init(
                dsn=os.getenv('SENTRY_DSN_URL'),
                environment=self.ENVIRONMENT,
                integrations=[
                    TornadoIntegration(),
                ],
                traces_sample_rate=1.0,
                release=f"{self.SERVICE_NAME}@{os.getenv('GITLAB_COMMIT')}" if os.getenv('GITLAB_COMMIT') else None,
            )

        if os.getenv("DD_SERVICE") and os.getenv("DD_API_KEY") and os.getenv("DD_API_KEY") and os.getenv("DD_ENV"):
            self.logging.info(
                f"[PLUGINS] DataDog enabled (service: {os.getenv('DD_SERVICE')}, env: {os.getenv('DD_ENV')})"
            )
            self.monitor = DataDogMonitor(
                host=os.getenv("DD_HOST", '127.0.0.1'),
                service=os.getenv("DD_SERVICE"),
                env=os.getenv("DD_ENV"),
                api_key=os.getenv("DD_API_KEY"),
                app_key=os.getenv("DD_APP_KEY"),
            )

        host = os.getenv('SERVICE_HOST', '0.0.0.0')
        if not kwargs.get('bind_port'):
            try:
                port = int(os.getenv('SERVICE_PORT', 50000))
            except (ValueError, TypeError):
                port = 50000
        else:
            port = kwargs.get('bind_port')

        self.logging.info(f"Starting {self.NAME.capitalize()} service on [{host}:{port}]")
        self.listen(address=host, port=port)
        for queue in queues:
            self.io_loop.add_callback(queue)
        try:
            self.io_loop.start()
        except KeyboardInterrupt:
            self.logging.info(f"Service {self.NAME.capitalize()} stopped")
            exit(0)


class Endpoint(web.RequestHandler):
    WORKING_DIRECTORY = None
    ENVIRONMENT = None

    logging = None
    redis = None
    mongo = None
    monitor = None

    def initialize(self):
        self.logging = self.application.logging
        self.WORKING_DIRECTORY = self.application.WORKING_DIRECTORY
        self.ENVIRONMENT = self.application.ENVIRONMENT
        self.redis = self.application.redis
        self.mongo = self.application.mongo
        self.monitor = self.application.monitor

    def set_default_headers(self) -> None:
        if os.getenv("CORS_ENABLED"):
            if os.getenv("CORS_ALLOWED_ORIGINS") == '*':
                self.set_header("Access-Control-Allow-Origin", "*")
            elif os.getenv("CORS_ALLOWED_ORIGINS"):
                allowed_origins = os.getenv("CORS_ALLOWED_ORIGINS").split(',')
                if self.request.headers.get('origin') in allowed_origins:
                    self.set_header("Access-Control-Allow-Origin", self.request.headers.get('origin'))

            self.set_header("Access-Control-Allow-Headers", "Content-Type, Depth, User-Agent, X-File-Size, "
                                                            "X-Requested-With, X-Requested-By, If-Modified-Since, "
                                                            "X-File-Name, Cache-Control, Authorization")
            if hasattr(self, 'methods') and isinstance(self.methods, list):
                self.methods.append('OPTIONS')
                self.set_header('Access-Control-Allow-Methods', ', '.join(self.methods))

    @property
    def user_ip(self):
        ip = self.request.headers.get('X-Real-Ip')
        if not ip:
            ip = self.request.remote_ip
        return ip

    @property
    def user_agent(self):
        return self.request.headers.get('User-Agent')

    @property
    def hostname(self):
        host = self.request.headers.get('origin')
        if not host:
            host = self.request.host

        if host:
            host = host.replace('https', '').replace('http', '').replace('://', '')
            if ':' in host:
                host = host.split(':')[0]
            return host
        return None

    @property
    def current_page(self):
        try:
            page = int(self.get_argument('page', 1)) - 1
        except ValueError:
            page = 0
        return page

    @property
    def pagination(self):
        limit = os.getenv("DB_RESULTS_LIMIT", 20)
        return {'limit': limit, 'offset': self.current_page * limit}

    def finish_with_error(self, status_code, code=None, fields=None):
        title = 'Unknown error'
        if status_code == 500:
            title = 'Internal Server Error'
        elif status_code == 400:
            title = 'Bad Request'
        elif status_code == 401:
            title = 'Unauthorized'
        elif status_code == 403:
            title = 'Forbidden'
        elif status_code == 404:
            title = 'Not Found'
        elif status_code == 405:
            title = 'Method Not Allowed'

        message = {
            'title': title,
            'status': status_code,
            'instance': self.request.path,
        }

        if code:
            message['code'] = code

        if fields:
            message['fields'] = fields

        self.set_header('Content-type', 'application/problem+json')
        self.set_status(status_code)
        return self.finish(message)

    def finish_with_pagination(self, result, total_count=0, parameters=None, limit=None):
        try:
            page = int(self.get_argument('page', 1)) - 1
        except ValueError:
            page = -1

        if limit is None:
            limit = os.getenv("DB_RESULTS_LIMIT", 20)

        count = 0
        pages = 0
        next_page = None
        prev_page = None
        if limit and len(result) and page >= 0 and total_count:
            count = total_count
            pages = int(count / limit)
            if count % limit != 0:
                pages += 1

            request_url = self.request.path
            if request_url[-1] != '/':
                request_url += '/'

            if (page + 1) < pages:
                next_page = f"{request_url}?page={page + 2}"
            if page > 0:
                prev_page = f"{request_url}?page={page}"

        if parameters:
            for name, value in parameters.items():
                if value:
                    if next_page:
                        next_page += f"&{name}={value}"
                    if prev_page:
                        prev_page += f"&{name}={value}"

        for i, s in enumerate(result):
            if isinstance(s, dict):
                for key, value in s.items():
                    if isinstance(value, datetime.date):
                        s[key] = value.strftime("%Y-%m-%d")
                    if isinstance(value, decimal.Decimal):
                        s[key] = float(value)
                continue
            result[i] = s.frontend

        return self.finish({
            'status': 200,
            'title': 'OK',
            'count': count if count else len(result),
            'size': limit if total_count else 0,
            'pages': pages,
            'next': next_page,
            'previous': prev_page,
            'results': result,
        })

    def finish_with_ok_status(self):
        self.set_header('Content-type', 'application/json')
        self.set_status(200)
        return self.finish({'ok': True})

    def finish(self, chunk: Optional[Union[str, bytes, dict]] = None) -> "Future[None]":
        if hasattr(chunk, 'frontend'):
            chunk = chunk.frontend
        elif isinstance(chunk, list):
            result = {'data': []}
            for row in chunk:
                if hasattr(row, 'frontend'):
                    result['data'].append(row.frontend)
                else:
                    result['data'].append(row)
            chunk = result
        return super().finish(chunk)

    def get_argument(self, name: str, default=None, strip: bool = True, argument_type=None, required: bool = False,
                     process=None, possible_values=None, validation=None, fields=None, argument_format=None):
        try:
            json_data = escape.json_decode(self.request.body)
        except (json.JSONDecodeError, UnicodeDecodeError):
            json_data = {}

        result = None

        if json_data.get(name) is not None:
            if str(json_data.get(name)).lower() == 'true':
                result = True
            elif str(json_data.get(name)).lower() == 'false':
                result = False
            else:
                result = json_data.get(name)
        elif super().get_argument(name, None) is not None:
            value = super().get_argument(name)
            if str(value).lower() == 'true':
                result = True
            elif str(value).lower() == 'false':
                result = False
            else:
                result = value

        if required:
            if result is None:
                raise ValidationException(field=name, message="Field is required")

        if result is not None and argument_type:
            if argument_type == int:
                try:
                    result = int(result)
                except (TypeError, ValueError, AttributeError):
                    raise ValidationException(field=name, message="Invalid numeric value")
            elif argument_type == float:
                try:
                    result = float(result)
                except (TypeError, ValueError, AttributeError):
                    raise ValidationException(field=name, message="Invalid decimal value")
            elif argument_type == str:
                try:
                    result = str(result)
                except (TypeError, ValueError, AttributeError):
                    raise ValidationException(field=name, message="Invalid string")
            elif argument_type == datetime.datetime:
                try:
                    if not argument_format:
                        argument_format = '%Y-%m-%dT%H:%M:%SZ%z'
                    result = datetime.datetime.strptime(result, argument_format)
                except (TypeError, ValueError, AttributeError):
                    raise ValidationException(field=name, message="Invalid datetime value")
            elif argument_type == datetime.date:
                try:
                    if not argument_format:
                        argument_format = '%Y-%m-%d'
                    result = datetime.datetime.strptime(result, argument_format).date()
                except (TypeError, ValueError, AttributeError):
                    raise ValidationException(field=name, message="Invalid date value")
            elif argument_type == list and not isinstance(result, list):
                raise ValidationException(field=name, message="Invalid array value")
            elif argument_type == bool and not isinstance(result, bool):
                raise ValidationException(field=name, message="Invalid boolean value")
            elif argument_type == dict and not isinstance(result, dict):
                raise ValidationException(field=name, message="Invalid object value")
            elif argument_type == 'uuid':
                try:
                    uuid.UUID(result, version=4)
                except (TypeError, ValueError, AttributeError):
                    raise ValidationException(field=name, message="Invalid UUID value")
            elif argument_type == 'email':
                regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'
                if not re.fullmatch(regex, result):
                    raise ValidationException(field=name, message="Invalid email")

        if required and not result:
            raise ValidationException(field=name, message="Field is required and should not be empty")

        if argument_type in ['email']:
            if not process:
                process = []
            process.extend(['lower', 'strip'])

        if result is not None and argument_type is str and process:
            if 'capitalize_first' in process:
                result = result.capitalize()
            if 'capitalize' in process:
                spaces_split = result.split(' ')
                for index, row in enumerate(spaces_split):
                    spaces_split[index] = str(row).capitalize()
                result = ' '.join(spaces_split)
            if 'strip' in process:
                result = result.strip()
            if 'lower' in process:
                result = result.lower()
            if 'upper' in process:
                result = result.upper()
            if 'phone' in process:
                result = result.replace(' ', '').replace('-', '').replace('(', '').replace(')', '').replace('+', '')

        if result is not None and possible_values and result not in possible_values:
            raise ValidationException(field=name, message="Invalid value")

        if result is not None and validation:
            for validate, parameters in validation.items():
                if validate == 'length':
                    if 'min' in parameters.keys():
                        if len(result) < parameters['min']:
                            raise ValidationException(
                                field=name, message=f"Value length too small (min: {parameters['min']})"
                            )
                    if 'max' in parameters.keys():
                        if len(result) > parameters['max']:
                            raise ValidationException(
                                field=name, message=f"Value length too big (max: {parameters['max']})"
                            )
                if validate == 'contains':
                    if 'upper' in parameters and not any(char.isupper() for char in result):
                        raise ValidationException(field=name, message="Value should contain upper letter")
                    if 'number' in parameters and not any(char.isdigit() for char in result):
                        raise ValidationException(field=name, message="Value should contain number")

        if result is not None and process and 'encode' in process:
            result = result.encode('utf-8')

        if result is not None:
            return result

        return default

    def get_arguments(self, arguments: dict) -> dict:
        result = {}
        exceptions = []
        for argument, parameters in arguments.items():
            try:
                result[argument] = self.get_argument(name=argument, default=parameters.get('default'),
                                                     argument_type=parameters.get('type'),
                                                     required=parameters.get('required'),
                                                     process=parameters.get('process'),
                                                     possible_values=parameters.get('possible_values'),
                                                     validation=parameters.get('validate'),
                                                     fields=parameters.get('fields'),
                                                     argument_format=parameters.get('format'))
            except ValidationException as e:
                exceptions.append(e)
        if exceptions:
            raise ManyValidationExceptions(exceptions=exceptions)
        return result

    @staticmethod
    def update_errors(error_fields, key, error):
        if key not in error_fields:
            error_fields[key] = [error]
        else:
            error_fields[key].append(error)
        return error_fields

    def get_file(self, attribute_name):
        file_handler = self.request.files.get(attribute_name)
        if not file_handler:
            return None
        file_handler = file_handler[0]

        mime = magic.Magic(mime=True)
        file_type = mime.from_buffer(file_handler.get('body'))

        result = {
            "file": file_handler.get('body'),
            "type": file_type,
            "name": file_handler.get('filename'),
            "extension": os.path.splitext(file_handler.get('filename'))[1]
        }

        return result

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    async def options(self, *args, **kwargs):
        if hasattr(self, 'methods') and isinstance(self.methods, list):
            self.set_header('Allow', ', '.join(self.methods))
            self.set_status(204)
            return self.finish()
        return self.finish_with_error(404)

    async def get(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def post(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def put(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def patch(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def delete(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def copy(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def head(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def link(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def unlink(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def purge(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def lock(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def unlock(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def propfind(self, *args, **kwargs):
        return self.finish_with_error(405)

    async def view(self, *args, **kwargs):
        return self.finish_with_error(405)


class HealthCheckEndpoint(Endpoint):
    methods = ['GET']

    def _log(self) -> None:
        return

    async def get(self, *args, **kwargs):
        return self.finish_with_ok_status()



import datetime
import os
import random
import re
import string
from decimal import Decimal
from uuid import uuid4

import aiopg
import psycopg2
import psycopg2.extras

SYSTEM_ATTRIBUTES = ('TABLE_NAME', 'FRONTEND_EXCLUDE_FIELDS', 'SPECIAL_FIELDS', 'COMMON_FRONTEND_EXCLUDE_FIELDS',
                     'frontend', 'COMMON_FRONTEND_INCLUDE_FIELDS', 'FRONTEND_INCLUDE_FIELDS')


class DatabaseRecordNotFoundError(Exception):
    pass


class DatabaseError(Exception):
    pass


class Relationship:
    def __init__(self, relation_class, relation_field, key='id', parameters=None):
        self.relation = relation_class
        self.field = relation_field
        self.key = key
        self.parameters = parameters


class Relationships:
    def __init__(self, relation_class, relation_field, key='id', parameters=None, sort_column=None, sort_order='asc'):
        self.relation = relation_class
        self.field = relation_field
        self.key = key
        self.parameters = parameters
        self.sort_column = sort_column
        self.sort_order = sort_order


class RelationCount:
    def __init__(self, relation_class, relation_field, key='id', parameters=None):
        self.relation = relation_class
        self.field = relation_field
        self.key = key
        self.parameters = parameters


class QueryField:
    def __init__(self, query, parameters=None):
        self.query = query
        self.parameters = parameters


class Model:
    COMMON_FRONTEND_EXCLUDE_FIELDS = ('id', 'is_active', 'is_deleted', 'created', 'modified', )
    COMMON_FRONTEND_INCLUDE_FIELDS = ()
    id = 0
    uuid = None
    is_active = True
    is_deleted = False
    created = None
    modified = None

    def __init__(self, **kwargs):
        for attribute in kwargs.keys():
            if hasattr(self, attribute):
                setattr(self, attribute, kwargs.get(attribute))
        if self.uuid is None:
            self.uuid = str(uuid4())
        if self.created is None:
            self.created = datetime.datetime.utcnow()
        if self.modified is None:
            self.modified = datetime.datetime.utcnow()

    @classmethod
    def table_short_name(cls):
        name = f"{cls.TABLE_NAME}"
        name = name.lower().replace('incase.', '')
        return name.strip()

    @classmethod
    def table_full_name(cls):
        return f"{cls.TABLE_NAME} {cls.table_short_name()}"

    @classmethod
    def attributes(cls, exclude_fields=()):
        attributes = []
        for attribute in dir(cls):
            if attribute in exclude_fields:
                continue
            if attribute in SYSTEM_ATTRIBUTES:
                continue
            if attribute.startswith('__'):
                continue
            if isinstance(getattr(cls, attribute), Relationship) \
                    or isinstance(getattr(cls, attribute), Relationships) \
                    or isinstance(getattr(cls, attribute), RelationCount):
                continue
            if callable(getattr(cls, attribute)):
                continue
            attributes.append(attribute)
        return sorted(attributes)

    @classmethod
    def sql_fields(cls, exclude_fields=(), db_exec=False):
        fields = []
        for attribute in cls.attributes(exclude_fields):
            if hasattr(cls, 'SPECIAL_FIELDS') and attribute in cls.SPECIAL_FIELDS:
                continue
            value = getattr(cls, attribute)
            if isinstance(value, QueryField) and not db_exec:
                fields.append(f'({value.query}) AS "{cls.table_short_name()}__{attribute}"')
                continue
            elif isinstance(value, QueryField) and db_exec:
                continue
            if db_exec:
                fields.append(f'"{attribute}"')
            else:
                fields.append(f'{cls.table_short_name()}.{attribute} AS "{cls.table_short_name()}__{attribute}"')
        return fields

    @classmethod
    def sql_format_fields(cls, exclude_fields=()):
        fields = []
        for attribute in cls.attributes(exclude_fields):
            value = getattr(cls, attribute)
            if hasattr(cls, 'SPECIAL_FIELDS') and attribute in cls.SPECIAL_FIELDS:
                continue
            if isinstance(value, QueryField):
                continue
            fields.append(f"%({attribute})s")
        return ", ".join(fields)

    @staticmethod
    def _process_parameter(table, field, value):
        field_name = f"{field}_{''.join(random.choice(string.ascii_lowercase) for _ in range(6))}".replace('::', '_')
        sql_name = f'{table.table_short_name()}.{field}'
        try:
            if isinstance(getattr(table, field), QueryField):
                sql_name = f'{table.table_short_name()}__{field}'
        except AttributeError:
            if '::' not in field and '->' not in field and 'value between' in field.lower():
                return [], []

        if isinstance(value, str):
            if value.lower() == 'is not null':
                return [f"{sql_name} IS NOT NULL"], []
            if value.lower() == 'is null':
                return [f"{sql_name} IS NULL"], []
            if 'now()' in value.lower():
                add = ''
                if '::date' in value.lower():
                    add = '::date'
                if 'interval' in value.lower():
                    sign = '+'
                    if '-' in value:
                        sign = '-'
                    add = f" {sign} {value.lower()[value.index('interval'):]}"
                if value.startswith('>='):
                    return [f"{sql_name} >= NOW(){add}"], []
                elif value.startswith('<='):
                    return [f"{sql_name} <= NOW(){add}"], []
                elif value.startswith('>'):
                    return [f"{sql_name} > NOW(){add}"], []
                elif value.startswith('<'):
                    return [f"{sql_name} < NOW(){add}"], []
                elif value.startswith('!='):
                    return [f"{sql_name} <> NOW(){add}"], []
                elif value.startswith('='):
                    return [f"{sql_name} = NOW(){add}"], []
                else:
                    return [], []
            if value.startswith("!="):
                return [f"{sql_name} != %({field_name})s"], [{field_name: value.replace('!=', '').strip()}]
            if value.startswith('>='):
                return [f"{sql_name} >= %({field_name})s"], [{field_name: value.replace('>=', '').strip()}]
            if value.startswith('<='):
                return [f"{sql_name} <= %({field_name})s"], [{field_name: value.replace('<=', '').strip()}]
            if value.startswith('>'):
                return [f"{sql_name} > %({field_name})s"], [{field_name: value.replace('>', '').strip()}]
            if value.startswith('<'):
                return [f"{sql_name} < %({field_name})s"], [{field_name: value.replace('<', '').strip()}]
            if value.lower().startswith('like'):
                return [f"{sql_name} LIKE %({field_name})s"], [{field_name: value[4:].replace("'", '').strip()}]
            if value.lower().startswith('ilike'):
                return [f"{sql_name} ILIKE %({field_name})s"], [{field_name: value[5:].replace("'", '').strip()}]
            if 'value between' in value.lower():
                values = value.split(' ')
                return [f"%({field_name})s BETWEEN {table.table_short_name()}.{values[3]} "
                        f"AND {table.table_short_name()}.{values[4]}"], [{field_name: values[0]}]
            if 'between' in value.lower():
                values = value.split(' ')
                return [f"{sql_name} BETWEEN %({field_name}_1)s AND %({field_name}_2)s"], \
                       [{f"{field_name}_1": values[1]}, {f"{field_name}_2": values[2]}]
            if 'query=[' in value.lower() or 'query in [' in value.lower():
                query = " ".join([line.strip() for line in value.splitlines()])
                query = re.search(r'\[(.*?)]', query).group(1).strip()
                if value.lower().startswith('query=['):
                    return [f"{sql_name} = ({query})"], []
                elif value.lower().startswith('query in ['):
                    return [f"{sql_name} IN ({query})"], []
                else:
                    return [], []
            if 'query!=[' in value.lower() or 'query not in [' in value.lower():
                query = " ".join([line.strip() for line in value.splitlines()])
                query = re.search(r'\[(.*?)]', query).group(1).strip()
                if value.lower().startswith('query!=['):
                    return [f"{sql_name} != ({query})"], []
                elif value.lower().startswith('query not in ['):
                    return [f"{sql_name} NOT IN ({query})"], []
                else:
                    return [], []
        if isinstance(value, tuple):
            fields = []
            values = []
            for item in value:
                f, v = Model._process_parameter(table, field, item)
                fields.append(f)
                values.append(v)
            return [f"({' OR '.join(fields)})"], values
        if isinstance(value, list):
            return [f'{sql_name} IN ({", ".join([str(e) for e in value])})'], []
        return [f"{sql_name} = %({field_name})s"], [{field_name: value}]

    @classmethod
    def _process_parameters(cls, parameters, table_class=None):
        if not parameters:
            return "", {}

        if table_class is None:
            table_class = cls

        fields = []
        result_parameters = {}
        if isinstance(parameters, dict):
            parameters = [parameters]

        for parameter in parameters:
            if isinstance(parameter, dict):
                for name, value in parameter.items():
                    f, values = table_class._process_parameter(table_class, name, value)
                    fields.extend(f)
                    for v in values:
                        result_parameters.update(v)
            if isinstance(parameter, list):
                or_fields = []
                for row in parameter:
                    for name, value in row.items():
                        f, values = table_class._process_parameter(table_class, name, value)
                        or_fields.extend(f)
                        for v in values:
                            result_parameters.update(v)
                if or_fields:
                    fields.append(f"({' OR '.join(or_fields)})")

        query = ''
        if fields:
            query = ' AND '.join(fields)
        return query, result_parameters

    @classmethod
    def process_relationships(cls):
        fields = []
        parameters = {}
        for attribute in dir(cls):
            value = getattr(cls, attribute)
            if not isinstance(value, Relationship) and not isinstance(value, Relationships) \
                    and not isinstance(value, RelationCount):
                continue
            rel = value.relation
            if isinstance(value, RelationCount):
                query = f"SELECT COUNT(*) FROM {rel.table_full_name()} "
                query += f"WHERE {rel.table_short_name()}.{value.field} = {cls.table_short_name()}.{value.key}"
                where, params = cls._process_parameters(value.parameters, table_class=rel)
                if where:
                    query += f" AND {where}"
                    parameters.update(params)
                fields.append(f"({query}) AS count__{cls.table_short_name()}__{attribute}")
                continue

            rel_fields = []
            for rel_attr in rel.attributes():
                if hasattr(rel, 'SPECIAL_FIELDS') and rel_attr in rel.SPECIAL_FIELDS:
                    continue
                if isinstance(getattr(rel, rel_attr), QueryField):
                    rel_fields.append(f"'{rel.table_short_name()}__{rel_attr}', "
                                      f"({getattr(rel, rel_attr).query})")
                else:
                    rel_fields.append(f"'{rel.table_short_name()}__{rel_attr}', {rel.table_short_name()}.{rel_attr}")
            rel_child_fields, params = rel.process_relationships()
            parameters.update(params)
            for rel_child_field in rel_child_fields:
                json_select = rel_child_field[:rel_child_field.rindex(' AS ')]
                json_name = rel_child_field[rel_child_field.rindex(' AS ')+4:]
                rel_fields.append(f"'{json_name}', {json_select}")

            if isinstance(value, Relationship):
                query = f"SELECT json_build_object({', '.join(rel_fields)}) "
            else:
                query = f"SELECT json_agg(json_build_object({', '.join(rel_fields)})) "
            query += f"FROM {rel.table_full_name()} "
            query += f"WHERE {rel.table_short_name()}.{value.field} = {cls.table_short_name()}.{value.key}"
            where, params = cls._process_parameters(value.parameters, table_class=rel)
            if where:
                query += f" AND {where}"
                parameters.update(params)
            fields.append(f"({query}) AS json__{cls.table_short_name()}__{attribute}")
        return fields, parameters

    @staticmethod
    async def fetch(query, values=None, many=False, expected_result=True, postgres_uri=None):
        if not postgres_uri:
            postgres_uri = os.getenv('DB_URL')

        try:
            pool = await aiopg.create_pool(postgres_uri, timeout=10)
            async with pool.acquire() as connection:
                async with connection.cursor(cursor_factory=psycopg2.extras.DictCursor) as cursor:
                    try:
                        await cursor.execute(query, values)
                    except Exception as exception:
                        await connection.close()
                        raise DatabaseError(exception)
                    if expected_result:
                        if not many:
                            result = await cursor.fetchone()
                            if result:
                                result = dict(result)
                        else:
                            result = []
                            rows = await cursor.fetchall()
                            for row in rows:
                                result.append(dict(row))
                await connection.close()
        except Exception as exception:
            raise DatabaseError(exception)

        if expected_result:
            return result

    @staticmethod
    def _process_record(cls, record):
        if not record:
            return None

        columns = {}
        for attribute in cls.attributes():
            if hasattr(cls, 'SPECIAL_FIELDS') and attribute in cls.SPECIAL_FIELDS:
                continue
            record_key = f"{cls.table_short_name()}__{attribute}"
            columns[attribute] = record.get(record_key)

        for attribute in dir(cls):
            value = getattr(cls, attribute)
            if not isinstance(value, Relationship) and not isinstance(value, Relationships) \
                    and not isinstance(value, RelationCount):
                continue
            rel = value.relation
            if isinstance(value, RelationCount):
                try:
                    columns[attribute] = record.get(f"count__{cls.table_short_name()}__{attribute}", 0)
                except (ValueError, TypeError):
                    columns[attribute] = 0
                continue
            json_record_key = f"json__{cls.table_short_name()}__{attribute}"
            json_record = record.get(json_record_key)
            if isinstance(value, Relationship):
                columns[attribute] = Model._process_record(rel, json_record)
            else:
                columns[attribute] = Model._process_records(rel, json_record)
                if value.sort_column:
                    columns[attribute] = sorted(
                        columns[attribute],
                        key=lambda d: getattr(d, value.sort_column),
                        reverse=True if value.sort_order == 'desc' else False
                    )

        return cls(**columns)

    @staticmethod
    def _process_records(cls, records):
        objects = []
        for record in records if records else []:
            objects.append(Model._process_record(cls, record))
        return objects

    @classmethod
    def _generate_sql(cls):
        fields = cls.sql_fields()
        f, parameters = cls.process_relationships()
        for attribute in cls.attributes():
            value = getattr(cls, attribute)
            if isinstance(value, QueryField) and value.parameters:
                for key, v in value.parameters.items():
                    parameters.update({key: v})
                continue

        fields.extend(f)
        query = f"SELECT {', '.join(fields)} FROM {cls.table_full_name()}"
        return query, parameters

    @classmethod
    async def get(cls, pk=0, parameters=None, active=True):
        if pk:
            try:
                pk = int(pk)
                parameters = {'id': pk}
            except (ValueError, TypeError):
                pass

        query, params = cls._generate_sql()
        where, parameters = cls._process_parameters(parameters)
        parameters.update(params)
        if where:
            query += f" WHERE {where}"
        if active:
            if not where:
                query += f" WHERE {cls.table_short_name()}.is_active = TRUE"
            else:
                query += f" AND {cls.table_short_name()}.is_active = TRUE"

        record = await cls.fetch(query, parameters)
        if not record:
            if pk:
                raise DatabaseRecordNotFoundError(f"Record id={pk} was not found")
            else:
                raise DatabaseRecordNotFoundError(f"Record with parameters {parameters} was not found")
        return cls._process_record(cls, record)

    @classmethod
    async def get_by_query(cls, query, parameters=None):
        if not hasattr(cls, 'TABLE_NAME'):
            raise ValueError(f"There is no TABLE_NAME attribute in class {cls.__name__}")

        record = await cls.fetch(query, parameters)
        if not record:
            return None
        return cls(**record)

    @classmethod
    async def all(cls, parameters=None, sort_column=None, sort_order='asc', limit=None, offset=None):
        query, params = cls._generate_sql()
        where, parameters = cls._process_parameters(parameters)
        parameters.update(params)
        if where:
            query += f" WHERE {where}"

        if sort_column:
            query += f" ORDER BY {cls.table_short_name()}__{sort_column} {sort_order}"
        if limit:
            query += f" LIMIT {limit}"
        if offset:
            query += f" OFFSET {offset}"

        records = await cls.fetch(query, parameters, many=True)
        return cls._process_records(cls, records)

    @classmethod
    async def count(cls, parameters=None, processed=None):
        if not parameters and not processed:
            query = f"SELECT COUNT(*) as count FROM {cls.table_full_name()}"
            count = await cls.fetch(query, parameters)
            if not count or not count.get('count'):
                return 0
            return count.get('count')

        if not processed:
            where, parameters = cls._process_parameters(parameters)
        else:
            where, parameters = processed

        query = f"SELECT COUNT(*) AS count FROM {cls.table_full_name()}"
        if where:
            query += f" WHERE {where}"

        count = await cls.fetch(query, parameters)
        if not count or not count.get('count'):
            return 0
        return count.get('count')

    @classmethod
    async def all_with_pagination(cls, parameters=None, sort_column=None, sort_order='asc', pagination=None):
        query, params = cls._generate_sql()
        where, parameters = cls._process_parameters(parameters)
        parameters.update(params)
        if where:
            query += f" WHERE {where}"

        count = await cls.count(processed=(where, parameters))
        if not count:
            return 0, []

        if sort_column:
            query += f" ORDER BY {cls.table_short_name()}__{sort_column} {sort_order}"
        if pagination:
            query += f" LIMIT {pagination.get('limit')} OFFSET {pagination.get('offset')} "

        records = await cls.fetch(query, parameters, many=True)
        return count, cls._process_records(cls, records)

    async def insert(self):
        if not hasattr(self, 'TABLE_NAME'):
            raise ValueError(f"There is no TABLE_NAME attribute in class {self.__name__}")

        fields = self.sql_fields(exclude_fields=('id',), db_exec=True)
        values = {}
        for field in fields:
            field = field.replace('"', '').strip()
            values[field] = getattr(self, field)

        query = f"INSERT INTO {getattr(self, 'TABLE_NAME')} " \
                f"({', '.join(fields)}) " \
                f"VALUES ({self.sql_format_fields(exclude_fields=('id',))}) RETURNING id"

        record_id = await self.fetch(query, values)
        self.id = record_id.get('id')

    async def update(self):
        if not hasattr(self, 'TABLE_NAME'):
            raise ValueError(f"There is no TABLE_NAME attribute in class {self.__name__}")

        self.modified = datetime.datetime.utcnow()

        fields = []
        values = {}
        for field in self.sql_fields(exclude_fields=('id', 'uuid', 'created', ), db_exec=True):
            field = field.replace('"', '').strip()
            values[field] = getattr(self, field)
            fields.append(f"{field} = %({field})s")

        values['id'] = self.id

        query = f"UPDATE {getattr(self, 'TABLE_NAME')} SET {', '.join(fields)} WHERE id = %(id)s"
        await self.fetch(query, values, expected_result=False)

    @classmethod
    async def delete(cls, pk=None, parameters=None):
        if not parameters:
            query = f"DELETE FROM {getattr(cls, 'TABLE_NAME')} WHERE id = %(id)s"
            await cls.fetch(query, {'id': pk}, expected_result=False)
            return

        where, parameters = cls._process_parameters(parameters)
        query = f"DELETE FROM {getattr(cls, 'TABLE_NAME')} {cls.table_short_name()}"
        if where:
            query += f" WHERE {where}"
        await cls.fetch(query, parameters, expected_result=False)

    @property
    def frontend(self):
        result = {}
        for attribute in dir(self):
            if attribute in SYSTEM_ATTRIBUTES:
                continue
            if attribute.startswith('__'):
                continue
            if callable(getattr(self, attribute)):
                continue
            if isinstance(getattr(self, attribute), RelationCount):
                result[attribute] = 0
                continue
            if isinstance(getattr(self, attribute), Relationship):
                result[attribute] = None
                continue
            if isinstance(getattr(self, attribute), Relationships):
                result[attribute] = []
                continue
            if isinstance(getattr(self, attribute), QueryField):
                result[attribute] = None
                continue

            include_fields = self.COMMON_FRONTEND_INCLUDE_FIELDS
            if hasattr(self, 'FRONTEND_INCLUDE_FIELDS'):
                include_fields += self.FRONTEND_INCLUDE_FIELDS
            exclude_fields = self.COMMON_FRONTEND_EXCLUDE_FIELDS
            if hasattr(self, 'FRONTEND_EXCLUDE_FIELDS'):
                exclude_fields += self.FRONTEND_EXCLUDE_FIELDS

            if attribute not in include_fields and attribute in exclude_fields:
                continue

            value = getattr(self, attribute)
            if isinstance(value, datetime.datetime):
                result[attribute] = value.strftime('%Y-%m-%dT%H:%M:%SZ')
            elif isinstance(value, datetime.date):
                result[attribute] = value.strftime('%Y-%m-%d')
            elif isinstance(value, Decimal):
                result[attribute] = float(value)
            elif isinstance(value, object) and hasattr(value, 'frontend'):
                result[attribute] = value.frontend
            elif isinstance(value, list):
                result_list = []
                for val in value:
                    if isinstance(val, object) and hasattr(val, 'frontend'):
                        result_list.append(val.frontend)
                    else:
                        result_list.append(val)
                result[attribute] = result_list
            else:
                result[attribute] = value
        return result

    def set_currency_field(self, field, value):
        try:
            value = int(Decimal("%0.2f" % value) * 100)
        except ValueError:
            raise ValueError(f"Invalid string value: got '{value}', expected decimal")

        if not value:
            value = 0
        setattr(self, field, value)



import datadog


class DataDogMonitor:
    def __init__(self, **kwargs):
        self.__service_name = kwargs.get('service')
        self.__env = kwargs.get('env')
        datadog.initialize(api_key=kwargs.get('api_key'), app_key=kwargs.get('app_key'), host_name=kwargs.get('host'))

    def duration(self, name, duration):
        datadog.statsd.histogram(
            f'{self.__service_name}.durations.{name}',
            duration,
            tags=['app:' + self.__service_name, 'env:' + self.__env]
        )

    def count(self, name):
        datadog.statsd.increment(
            f'{self.__service_name}.metric.{name}',
            tags=['app:' + self.__service_name, 'env:' + self.__env],
            value=1
        )



from motor.motor_tornado import MotorClient


class MongoDB:
    def __init__(self, **kwargs):
        self.url = kwargs.get('url')

    def __connect(self):
        if not self.url:
            raise Exception("MongoDB: no connection url specified")
        mongo = MotorClient(self.url)
        return mongo.get_default_database()

    async def insert(self, collection, data, remove=False):
        db = self.__connect()
        db_collection = db[collection]
        await db_collection.create_index(keys='_id', name=f'{collection}Index')
        if remove:
            await db_collection.delete_many({})
        if isinstance(data, list):
            await db_collection.insert_many(data)
        else:
            await db_collection.insert_one(data)

    async def get(self, collection, many=True, parameters=None):
        db = self.__connect()
        if not parameters:
            parameters = {}
        if many:
            count = await db[collection].count_documents(parameters)
            result = await db[collection].find(parameters).to_list(length=count)
        else:
            result = await db[collection].find_one(parameters)
        return result



from setuptools import setup, find_packages

VERSION = '0.0.47'
DESCRIPTION = 'Simple async API framework'
LONG_DESCRIPTION = """
This package contains simple async API framework, based on Tornado.

Sources can be found on GitHub: https://github.com/ihormaioroff/simple-api-framework
"""

setup(
    name="simple-api-framework",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="Ihor Maiorov",
    author_email="ihor.maioroff@gmail.com",
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'tornado>=6.2',
        'python-dotenv>=0.20.0',
        'python-magic>=0.4.27',
        'aioredis>=2.0.1',
        'aiopg>=1.3.4',
        'psycopg2-binary>=2.9.3',
        'motor>=3.0.0',
        'ddtrace>=1.4.3',
        'datadog>=0.44.0',
        'sentry-sdk>=1.9.4'
    ],
    keywords=['tornado', 'api', 'simple-api'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3",
    ]
)



import datetime
import decimal
import json
from itertools import islice

import aioredis


class Redis:
    def __init__(self, **kwargs):
        self.prefix = ''
        if kwargs.get('prefix'):
            self.prefix = f"{kwargs.get('prefix')}:"
        self.url = kwargs.get('url')
        self.max_insert_size = kwargs.get('max_insert_size', 50000)
        self.connect_timeout = kwargs.get('connect_timeout', 60)
        self.timeout = kwargs.get('timeout', 60)
        self.retries = kwargs.get('retries', 3)

    def __pre_process_new_data(self, data):
        if isinstance(data, dict):
            result_dict = {}
            for key, value in data.items():
                if isinstance(key, datetime.date):
                    key = key.strftime("%Y-%m-%d")
                result_dict[key] = self.__pre_process_new_data(value)
            data = result_dict.copy()
        elif isinstance(data, list):
            for index in range(len(data)):
                data[index] = self.__pre_process_new_data(data[index])
        elif isinstance(data, datetime.date):
            data = data.strftime("%Y-%m-%d")
        elif isinstance(data, decimal.Decimal):
            data = float(data)
        return data

    def __chunks(self, data):
        it = iter(data)
        for i in range(0, len(data), self.max_insert_size):
            yield {k: data[k] for k in islice(it, self.max_insert_size)}

    async def __connection(self):
        if not self.url:
            raise Exception("You need to set Redis URL")

        redis = await aioredis.from_url(
            self.url, encoding="utf-8", socket_connect_timeout=self.connect_timeout, socket_timeout=self.timeout
        )
        return redis

    async def __put(self, key, data, tries=0):
        redis = await self.__connection()

        for item in self.__chunks(data):
            try:
                await redis.hset(f"{self.prefix}{key}", mapping=item)
            except:
                tries += 1
                if tries > self.retries:
                    return
                return await self.__put(key, data, tries)

    async def __remove(self, key):
        redis = await self.__connection()
        await redis.execute_command("del", f"{self.prefix}{key}")

    async def set(self, key, dictionary, remove=False):
        try:
            data = self.__pre_process_new_data(dictionary)
            if isinstance(data, dict):
                for data_key, value in data.items():
                    data[data_key] = json.dumps(value)
            elif isinstance(data, list):
                data = json.dumps(data)
            else:
                raise Exception("Redis values can be instances of dict or list")
            if remove:
                await self.__remove(key)
            await self.__put(key, data)
        except Exception as e:
            raise e

    async def get(self, key, elements=None):
        try:
            redis = await self.__connection()
            if not elements:
                results = {}
                result = await redis.hgetall(f"{self.prefix}{key}")
                if result:
                    for k, v in result.items():
                        results[k] = json.loads(v)
                return results
            elif isinstance(elements, list) or isinstance(elements, tuple):
                results = {}
                pipeline = redis.pipeline()
                for e in elements:
                    pipeline.hget(f"{self.prefix}{key}", e)
                result = await pipeline.execute()
                if result:
                    for i, k in enumerate(elements):
                        if result[i]:
                            results[k] = json.loads(result[i])
                        else:
                            results[k] = None
                return results
            else:
                result = await redis.hget(f"{self.prefix}{key}", elements)
                if result:
                    result = json.loads(result)
                return result
        except Exception as e:
            raise e
